/*****************************************************************************
 *   Project:     GENETIC5.PRJ                                               *
 *****************************************************************************
 *   Modules:     GENPARM5.H                                                 *
 *                EXTERN5.H                                                  *
 *                INTERN5.H                                                  *
 *                GENETIC5.C                                                 *
 *                POPUL5.C                                                   *
 *                EVAL5.C                                                    *
 *                DISPLAY5.C                                                 *
 *                REPRO5.C                                                   *
 *****************************************************************************
 *   Topic:       DISPLAY5.C                                                 *
 *****************************************************************************
 *   Description: This module contains all functions and procedures that are *
 *                responsible for I/O. Keyboard-handlers aswell as text or   *
 *                graphical output are provided by this module.              *
 *****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <graphics.h>
#include <values.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "extern5.h"

/*****************************************************************************
 *   Function:    DISPLAY_RESULT                                             *
 *****************************************************************************
 *   Description: This function handles keystrokes by the user and calls the *
 *                propriate functions.                                       *
 *****************************************************************************
 *   Input:       GENERATION *gen - the generation to be handled             *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void display_result(GENERATION *gen)
{
   static unsigned char mode = TEXT;              /* currently active screen */

   if(kbhit())                                    /* key pressed ? */
   {
       switch(getch())                            /* which one is it ? */
       {
          case 't':                               /* display information of current generation on text screen */
          case 'T':mode = TEXT;
                   break;
          case 'g':                               /* display information of current generation graphically */
          case 'G':mode = GRAPHIC;
                   break;
          case 'd':                               /* show evolution details */
          case 'D':mode = DETAILS;
                   break;
          case 'f':                               /* display fitness progress of evolution */
          case 'F':mode = FITNESS;
                   break;
          case 27 :closegraph();                  /* restore default screen mode */
                   exit(0);                       /* terminate */
          default :mode = ++mode % EXIT;          /* show next screen */
                   break;
       }
       clrscr();                                  /* if mode changes clear screen */
   }
   switch(mode)                                   /* which mode is selected ? */
   {
       case TEXT:   text_display(gen);            /* textmode */
                    break;
       case GRAPHIC:graphic_display(gen);         /* graphic mode */
                    break;
       case DETAILS:detail_display(gen);          /* evolution details */
                    break;
       case FITNESS:fitness_display(gen->number); /* fitness chart */
                    break;
       default:     printf("\n\n%c%cUnknown screenmode !\n\n",7,7);
                    break;
   }
   if(LOG_MODE)
      write_file(gen);                            /* write information to file */
   if(PAUSE_MODE)
      getch();                                    /* wait for keystroke */
}


/*****************************************************************************
 *   Function:    GET_FORMULA                                                *
 *****************************************************************************
 *   Description: This function converts the genetic information to the      *
 *                represented formula for display purposes.                  *
 *****************************************************************************
 *   Input:       CHROMOSOME *ind - individuum to be converted               *
 *****************************************************************************
 *   Output:      char *txt - assembled text                                 *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void get_formula(char *txt,CHROMOSOME *ind)
{
   int i;                                 /* loop index */
   char tmp[40];                          /* temporary character string */

   txt[0] = '\0';                         /* initialize string */

   for(i=0;i<CHROMOSOME_SIZE;i+=2)        /* look at whole chromosome */
   {
    switch(ind->coded_str[i])             /* decode genetic code */
    {
         case IDENT:sprintf(tmp,"x,");
                    break;
         case ADD:  sprintf(tmp,"x+%lf,",(double)decode(ind->coded_str[i+1]));
                    break;
         case DIV:  sprintf(tmp,"x/%lf,",(double)decode(ind->coded_str[i+1]));
                    break;
         case SUB:  sprintf(tmp,"x-%lf,",(double)decode(ind->coded_str[i+1]));
                    break;
         case POW:  sprintf(tmp,"x^%lf,",(double)decode(ind->coded_str[i+1]));
                    break;
         case MULT: sprintf(tmp,"x*%lf,",(double)decode(ind->coded_str[i+1]));
                    break;
         case NEG:  sprintf(tmp,"-x,");
                    break;
         case LN:   sprintf(tmp,"ln(x),");
                    break;
         case EXP:  sprintf(tmp,"exp(x),");
                    break;
         case SIN:  sprintf(tmp,"sin(x),");
                    break;
         case COS:  sprintf(tmp,"cos(x),");
                    break;
         case SQRT: sprintf(tmp,"x,");
                    break;
         case RECI: sprintf(tmp,"1/x,");
                    break;
         default:  printf("\n\n%c%cWrong Operator Type !\n\n",7,7);
                   break;
    }
    strcat(txt,tmp);                       /* append current operator to formula */
  }
}


/*****************************************************************************
 *   Function:    TEXT_DISPLAY                                               *
 *****************************************************************************
 *   Description: This function displays all relevant information about the  *
 *                current generation in text mode.                           *
 *****************************************************************************
 *   Input:       GENERATION *gen - generation to be displayed               *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void text_display(GENERATION *gen)
{
   int i;                          /* index for population & parameters */
   double alpha1,                  /* decoded alpha */
          beta;                    /* decoded beta */
   char txt[256];

   restorecrtmode();               /* restore default textmode */
   gotoxy(1,1);

   /* display header: Generation number and total fitness */

   printf("Generation: %5d, Fitness: %10.5lf",gen->number,gen->total_fitness);
   printf("\n====================================================================");

   /* display each set of parameters (alpha,x,beta) of the best individuum */

   get_formula(txt,&gen->individuum[BEST]);
   printf("\nFormula:");
   printf("\n%s",txt);
   clreol();

   /* display the individual fitness of the best and worst chromosome and the average fitness */

   printf("\n\nFitness best: %10.5lf (generation), %10.5lf (overall)\n",gen->individuum[BEST].fitness,very_best_ind.fitness);
   printf("       worst: %10.5lf\n",gen->individuum[WORST].fitness);
   printf("     average: %10.5lf\n",gen->total_fitness/POP_SIZE);

   /* display the values of the function, the dest. function and the error in each evaluation point */

   for(i=1;i<=INPUT_ELEMENTS;++i)
   {
         alpha1 = function(&gen->individuum[BEST],(double)i);  /* value of function */
         beta   = dest_function((double)i);                    /* value of destination function */
         printf("\np(%2d) = %10.5lf      f(%2d) = %10.5lf      d(%2d) = %10.5lf",i,alpha1,i,beta,i,(alpha1 > beta) ? alpha1 - beta : beta - alpha1);
   }
}


/*****************************************************************************
 *   Function:    GRAPHIC_DISPLAY                                            *
 *****************************************************************************
 *   Description: This function displays the function and the destination    *
 *                function graphically.                                      *
 *****************************************************************************
 *   Input:       GENERATION *g - the generation to be displayed             *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void graphic_display(GENERATION *g)
{

    double i,                         /* loop index for double values */
           result,                    /* stores the result of both the function and the dest. function */
           minx,                      /* minimum x-value to be displayed */
           maxx,                      /* maximum x-value to be displayed */
           miny,                      /* minimum y-value to be displayed */
           maxy;                      /* maximum y-value to be displayed */
    int xcoor,                        /* x-coordinate of the pixel/line to draw */
        ycoor,                        /* y-coordinate of the pixel/line to draw */
        j,                            /* loop index for integer values */
        xpixels,                      /* number of horizontal pixels available on screen */
        ypixels;                      /* number of vertical pixels available on screen */
    char tmp[40];

    setgraphmode(getgraphmode());   /* restore default graphicmode */

    minx = MINX;                      /* set minimum x-value as user defined */
    maxx = MAXX;                      /* set maximum x-value as user defined */
    miny = MAXDOUBLE;                 /* initialize minimum y-value (largest positive) */
    maxy = -MAXDOUBLE;                /* initialize maximum y-value (largest negative) */
    xpixels = getmaxx();              /* determine horizontal resolution */
    ypixels = getmaxy();              /* determine vertical resolution */

    /* if autoscaling is desired, determine the smallest and largest y-coordinate */
    /* of the function that has to be displayed */

    for(i=minx;i<maxx && AUTOSCALE;i+=XSTEP)
    {
         if((result = function(&g->individuum[BEST],i)) > maxy)
            maxy = result;
         if(result < miny)
            miny = result;
     }

    /* determine smallest and largest y-coordinate of the dest. function to be displayed */

    for(i=minx;i<maxx;i+=XSTEP)
    {
        if((result = dest_function(i)) > maxy)
           maxy = result;
        if(result < miny)
           miny = result;
    }

    /* Draw vertical, dotted line at each evaluation point */

    setcolor(YELLOW);
    setlinestyle(DOTTED_LINE,0,NORM_WIDTH);
    settextstyle(SMALL_FONT,HORIZ_DIR,4);
    for(j=0;j<INPUT_ELEMENTS;++j)
    {
        xcoor = (int)((double)xpixels * (j + 1 - minx) / (maxx - minx));
        ycoor = 0;
        sprintf(tmp," %3.3lf",minx+j*(maxx-minx+1)/(double)INPUT_ELEMENTS);
        if(j > 0)
           outtextxy(xcoor,ycoor,tmp);
        moveto(xcoor,ycoor);
        lineto(xcoor,ypixels);
        sprintf(tmp,"- %3.3lf",maxy-j*(maxy-miny)/(double)INPUT_ELEMENTS);
        outtextxy(1,j * ypixels/INPUT_ELEMENTS,tmp);
    }

    /* scale and draw the blue destination function */

    setcolor(LIGHTBLUE);
    setlinestyle(SOLID_LINE,0,NORM_WIDTH);
    xcoor = 0;
    ycoor = (int)((double)ypixels - (ypixels * (dest_function(minx) - miny) / (maxy - miny)));
    moveto(xcoor,ycoor);

    for(i=minx+XSTEP;i<maxx;i+=XSTEP)
    {
        xcoor = (int)((double)xpixels * (i - minx) / (maxx - minx));
        ycoor = (int)((double)ypixels - (ypixels * (dest_function(i) - miny) / (maxy - miny)));
        lineto(xcoor,ycoor);
    }

    /* scale and draw the red function of the best of generation */

    if(g->individuum[BEST].fitness != very_best_ind.fitness)
    {
       setcolor(LIGHTRED);
       xcoor = 0;
       ycoor = (int)((double)ypixels - (ypixels * (function(&g->individuum[BEST],minx)-miny) / (maxy - miny)));
       moveto(xcoor,ycoor);

       for(i=minx+XSTEP;i<maxx;i+=XSTEP)
       {
           xcoor = (int)((double)xpixels * (i - minx) / (maxx - minx));
           ycoor = (int)((double)ypixels - (ypixels * (function(&g->individuum[BEST],i) - miny) / (maxy - miny)));
           lineto(xcoor,ycoor);
       }
    }

    /* scale and draw the green function of the overall best */

    setcolor(GREEN);
    xcoor = 0;
    ycoor = (int)((double)ypixels - (ypixels * (function(&very_best_ind,minx)-miny) / (maxy - miny)));
    moveto(xcoor,ycoor);

    for(i=minx+XSTEP;i<maxx;i+=XSTEP)
    {
        xcoor = (int)((double)xpixels * (i - minx) / (maxx - minx));
        ycoor = (int)((double)ypixels - (ypixels * (function(&very_best_ind,i) - miny) / (maxy - miny)));
        lineto(xcoor,ycoor);
    }

    setcolor(WHITE);
    settextstyle(SMALL_FONT,HORIZ_DIR,5);
    sprintf(tmp,"Generation: %3d",g->number);
    outtextxy(xpixels-115,ypixels-14,tmp);
}


void detail_display(GENERATION *g)
{
     restorecrtmode();
     gotoxy(1,1);

     printf("EVOLUTION DETAILS of generation %u",g->number);
     printf("\n=====================================================================");
     printf("\n    # evaluation points: %u",INPUT_ELEMENTS);
     printf("\n # operators of formula: %u",OPERATORS);
     printf("\ntrigonometric functions: %s",(TRIGONOMETRIC) ? "yes" : "no");
     printf("\n          operand range: [%lf,%lf]",LOWEST_OPERAND,LARGEST_OPERAND);
     printf("\n           operand type: %s",((OPERAND_TYPE)3/(OPERAND_TYPE)4 != 0) ? "double" : "int");
     printf("\n      chromosome length: %u",CHROMOSOME_SIZE);
     printf("\n     size of population: %u individuums",POP_SIZE);
     printf("\n  # individuums created: %ld of %ld",(long)g->number * (long)POP_SIZE,TRIALS);
     printf("\n\n  crossover probability: %f",CROSS_OVER_RATE);
     printf("\n   mutation probability: %f",MUTATION_RATE);
     printf("\n       replacement rate: %u",REPLACE_RATE);
     printf("\n           elitism rate: %u",ELITISM_RATE);
     printf("\n\n           fitness type: %s",str_fitness);
     printf("\n       parent selection: %s",str_selection);
     printf("\n\n               log mode: %s",(LOG_MODE) ? "on" : "off");
     printf("\n             pause mode: %s",(PAUSE_MODE) ? "on" : "off");
     printf("\n            autoscaling: %s",(AUTOSCALE) ? "on" : "off");
}

/*****************************************************************************
 *   Function:    WRITE_FILE                                                 *
 *****************************************************************************
 *   Description: This function creates a logfile and protocols all          *
 *                generational information into it (for later statistics)    *
 *****************************************************************************
 *   Input:       GENERATION *gen - the generation to be logged              *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void write_file(GENERATION *gen)
{
   int i;                          /* index for population & parameters */
   double alpha1,                  /* decoded alpha */
          beta;                    /* decoded beta */
   char txt[256];
   static FILE *fp;

   if(gen->number == 1)            /* if it is the first generation */
   {
     if((fp = fopen("GENETIC5.LOG","wb")) == NULL)   /* create logfile */
     {
         printf("\n\n%c%cError opening file 'GENETIC5.LOG'!\n\n",7,7);
         closegraph();
         exit(1);
     }

     /* Write details of evolution process */

     fprintf(fp,"=====================================================================");
     fprintf(fp,"\nEVOLUTION DETAILS ");
     fprintf(fp,"\n=====================================================================");
     fprintf(fp,"\n    # evaluation points: %u",INPUT_ELEMENTS);
     fprintf(fp,"\n # operators of formula: %u",OPERATORS);
     fprintf(fp,"\ntrigonometric functions: %s",(TRIGONOMETRIC) ? "yes" : "no");
     fprintf(fp,"\n          operand range: [%lf,%lf]",LOWEST_OPERAND,LARGEST_OPERAND);
     fprintf(fp,"\n           operand type: %s",((OPERAND_TYPE)3/(OPERAND_TYPE)4 != 0) ? "double" : "int");
     fprintf(fp,"\n      chromosome length: %u",CHROMOSOME_SIZE);
     fprintf(fp,"\n     size of population: %u individuums",POP_SIZE);
     fprintf(fp,"\n# individuums to create: %ld",TRIALS);
     fprintf(fp,"\n\n  crossover probability: %f",CROSS_OVER_RATE);
     fprintf(fp,"\n   mutation probability: %f",MUTATION_RATE);
     fprintf(fp,"\n       replacement rate: %u",REPLACE_RATE);
     fprintf(fp,"\n           elitism rate: %u",ELITISM_RATE);
     fprintf(fp,"\n\n           fitness type: %s",str_fitness);
     fprintf(fp,"\n       parent selection: %s",str_selection);
   }

   /* display header: Generation number and total fitness */

   fprintf(fp,"\n\n====================================================================");
   fprintf(fp,"\nGeneration: %5d, Fitness: %10.5lf",gen->number,gen->total_fitness);
   fprintf(fp,"\n====================================================================");

   /* display each set of parameters (alpha,x,beta) of the best individuum */

   get_formula(txt,&gen->individuum[BEST]);
   fprintf(fp,"\nFormula:");
   fprintf(fp,"\n%s",txt);

   /* display the individual fitness of the best and worst chromosome and the average fitness */

   fprintf(fp,"\n\nFitness best: %10.5lf (generation), %10.5lf (overall)\n",gen->individuum[BEST].fitness,very_best_ind.fitness);
   fprintf(fp,"       worst: %10.5lf\n",gen->individuum[WORST].fitness);
   fprintf(fp,"     average: %10.5lf\n",gen->total_fitness/POP_SIZE);

   /* display the values of the function, the dest. function and the error in each evaluation point */

   for(i=1;i<=INPUT_ELEMENTS;++i)
   {
         alpha1 = function(&gen->individuum[BEST],(double)i);  /* value of function */
         beta   = dest_function((double)i);                    /* value of destination function */
         fprintf(fp,"\np(%2d) = %10.5lf      f(%2d) = %10.5lf      d(%2d) = %10.5lf",i,alpha1,i,beta,i,(alpha1 > beta) ? alpha1 - beta : beta - alpha1);
   }
}

/*****************************************************************************
 *   Function:    FITNESS_DISPLAY                                            *
 *****************************************************************************
 *   Description: This function displays a chart of the fitness of all       *
 *                populations during the evolution.                          *
 *****************************************************************************
 *   Input:       GENERATION *gen - the generation to be displayed           *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/
void fitness_display(int gen)
{

    double bv,
           miny,                      /* minimum y-value to be displayed */
           maxy;                      /* maximum y-value to be displayed */
    int xcoor,                        /* x-coordinate of the pixel/line to draw */
        ycoor,                        /* y-coordinate of the pixel/line to draw */
        bx,
        by,
        bg,
        i,                            /* loop index for integer values */
        minx,                         /* minimum x-value to be displayed */
        maxx,                         /* maximum x-value to be displayed */
        xpixels,                      /* number of horizontal pixels available on screen */
        ypixels;                      /* number of vertical pixels available on screen */
    char txt[40];

    if(gen == 1)
       return;

    setgraphmode(getgraphmode());   /* restore default graphicmode */
    setlinestyle(SOLID_LINE,0,NORM_WIDTH);
    settextstyle(SMALL_FONT,HORIZ_DIR,4);

    minx = 0;                         /* set minimum x-value as user defined */
    maxx = gen;                       /* set maximum x-value as user defined */
    miny = 0.0;                       /* initialize minimum y-value */
    maxy = very_best_ind.fitness;     /* initialize maximum y-value */
    xpixels = getmaxx() - 30;         /* determine horizontal resolution */
    ypixels = getmaxy();              /* determine vertical resolution */

    /* Draw vertical, dotted line as y-axis */

    setcolor(YELLOW);
    setlinestyle(DOTTED_LINE,0,NORM_WIDTH);
    settextstyle(SMALL_FONT,HORIZ_DIR,4);
    for(i=0;i<5;++i)
    {
        xcoor = (int)(xpixels * i / 4);
        ycoor = 0;
        sprintf(txt," %3d",(int)((float)i*(float)gen/4. + .5));
        if(i > 0 && i < 4)
           outtextxy(xcoor,ycoor,txt);
        else if(i == 4)
           outtextxy(xcoor-25,ycoor,txt);
        moveto(xcoor,ycoor);
        lineto(xcoor,ypixels);
        sprintf(txt,"- %3.3lf",maxy-i*(maxy-miny)/5.0);
        outtextxy(1,i * ypixels/5,txt);
    }

    /* Display curve of the worst fitness per generation in blue */

    setcolor(LIGHTBLUE);
    setlinestyle(SOLID_LINE,0,NORM_WIDTH);
    bv = -MAXDOUBLE;
    xcoor = 0;
    ycoor = (int)((double)ypixels - (ypixels * (fitness_chart[minx].worst_fitness - miny) / (maxy - miny)));
    moveto(xcoor,ycoor);

    for(i=minx+1;i<maxx;i++)
    {
        xcoor = (int)((double)xpixels * (i - minx) / (maxx - minx - 1));
        ycoor = (int)((double)ypixels - (ypixels * (fitness_chart[i].worst_fitness - miny) / (maxy - miny)));
        lineto(xcoor,ycoor);
        if(fitness_chart[i].worst_fitness > bv)
        {
             bx = xcoor;
             by = ycoor;
             bg = i+1;
             bv = fitness_chart[i].worst_fitness;
        }
    }
    sprintf(txt,"%3.3lf",fitness_chart[gen-1].worst_fitness);
    outtextxy(xcoor+3,ycoor-12,txt);
    if(bv != fitness_chart[gen-1].worst_fitness)
    {
       sprintf(txt,"%3.3lf/%d",bv,bg);
       outtextxy(bx,by-12,txt);
    }

    /* Display curve of the average fitness per generation in red */

    setcolor(LIGHTRED);
    bv = -MAXDOUBLE;
    xcoor = 0;
    ycoor = (int)((double)ypixels - (ypixels * (fitness_chart[minx].average_fitness - miny) / (maxy - miny)));
    moveto(xcoor,ycoor);

    for(i=minx+1;i<maxx;i++)
    {
        xcoor = (int)((double)xpixels * (i - minx) / (maxx - minx - 1));
        ycoor = (int)((double)ypixels - (ypixels * (fitness_chart[i].average_fitness - miny) / (maxy - miny)));
        lineto(xcoor,ycoor);
        if(fitness_chart[i].average_fitness > bv)
        {
             bx = xcoor;
             by = ycoor;
             bg = i+1;
             bv = fitness_chart[i].average_fitness;
        }
    }
    sprintf(txt,"%3.3lf",fitness_chart[gen-1].average_fitness);
    outtextxy(xcoor+3,ycoor,txt);
    if(bv != fitness_chart[gen-1].average_fitness)
    {
       sprintf(txt,"%3.3lf/%d",bv,bg);
       outtextxy(bx,by,txt);
    }

    /* Display curve of the best fitness per generation in green */

    setcolor(GREEN);
    bv = -MAXDOUBLE;
    xcoor = 0;
    ycoor = (int)((double)ypixels - (ypixels * (fitness_chart[minx].best_fitness - miny) / (maxy - miny)));
    moveto(xcoor,ycoor);

    for(i=minx+1;i<maxx;i++)
    {
        xcoor = (int)((double)xpixels * (i - minx) / (maxx - minx - 1));
        ycoor = (int)((double)ypixels - (ypixels * (fitness_chart[i].best_fitness - miny) / (maxy - miny)));
        lineto(xcoor,ycoor);
        if(fitness_chart[i].best_fitness > bv)
        {
             bx = xcoor;
             by = ycoor;
             bg = i+1;
             bv = fitness_chart[i].best_fitness;
        }
    }
    sprintf(txt,"%3.3lf",fitness_chart[gen-1].best_fitness);
    outtextxy(xcoor+3,ycoor,txt);
    if(bv != fitness_chart[gen-1].best_fitness)
    {
       sprintf(txt,"%3.3lf/%d",bv,bg);
       outtextxy(bx,by,txt);
    }
}
