/*****************************************************************************
 *   Project:     GENETIC5.PRJ                                                *
 *****************************************************************************
 *   Modules:     GENPARM5.H                                                 *
 *                EXTERN5.H                                                  *
 *                INTERN5.H                                                  *
 *                GENETIC5.C                                                 *
 *                POPUL5.C                                                   *
 *                EVAL5.C                                                    *
 *                DISPLAY5.C                                                 *
 *                REPRO5.C                                                   *
 *****************************************************************************
 *   Topic:       GENETIC5.C                                                  *
 *****************************************************************************
 *   Description: This is the main module of the GENETIC Project. It         *
 *                contains the calling functions and other procedures that   *
 *                do not directly depend on or influence the GA, especially  *
 *                the evaluation procedures of the destination function.     *
 *****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <graphics.h>
#include <values.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "intern5.h"


/*****************************************************************************
 *   Function:    MATHERR                                                    *
 *****************************************************************************
 *   Description: This function is called in case of mathematical errors     *
 *                while computations.                                        *
 *****************************************************************************
 *   Input:       struct exception *e - structure containing error code,     *
 *                                      function, arguments and return value *
 *                                      causing the error                    *
 *****************************************************************************
 *   Output:      struct exception *e - structure containing error code,     *
 *                                      function, arguments and return value *
 *                                      causing the error                    *
 *****************************************************************************
 *   Return:      int matherr - 1 if error could be corrected                *
 *                              0 if not
 *****************************************************************************/

int matherr(struct exception *e)
{
   switch(e->type)      /* Which kind of error occured ? */
   {
       case DOMAIN:     printf("\n%cDOMAIN Fehler bei Berechnung von %s(%lf,%lf) !",7,e->name,e->arg1,e->arg2);
                        printf("\nFehler kann nicht korrigiert werden !");
                        getch();
                        return(0);
       case SING:       printf("\n%cSING Fehler bei Berechnung von %s(%lf,%lf) !",7,e->name,e->arg1,e->arg2);
                        printf("\nWert wird auf %lf gesetzt !",e->arg1);
                        e->retval = e->arg1;
                        getch();
                        return(1);
       case OVERFLOW:   e->retval = MAXDOUBLE;
                        return(1);
       case UNDERFLOW:  e->retval = MINDOUBLE;
                        return(1);
       case TLOSS:      printf("\n%cTLOSS Fehler bei Berechnung von %s(%lf,%lf) !",7,e->name,e->arg1,e->arg2);
                        printf("\nFehler kann nicht korrigiert werden !");
                        getch();
                        return(0);
       default:         printf("\n%cUNBEKANNTER Fehler bei Berechnung von %s(%lf,%lf) !",7,e->name,e->arg1,e->arg2);
                        printf("\nFehler kann nicht korrigiert werden !");
                        getch();
                        return(0);
   }
}


/*****************************************************************************
 *   Function:    EVALUATE_STACK                                             *
 *****************************************************************************
 *   Description: Uses a value, the code of an operator and an operand to    *
 *                compute a new value. Example: value = 9                    *
 *                                              decoded operator = ^         *
 *                                              operand = 3                  *
 *                                              new value = 729              *
 *****************************************************************************
 *   Input:       double *value          - value to be used as argument      *
 *                unsigned char operator - code of operator to be used       *
 *                double operand         - operand to be used for computation*
 *****************************************************************************
 *   Output:      double *value - new value computed                         *
 *****************************************************************************
 *   Return:      int evaluate_stack - 1 if computation is not valid         *
 *                                     0 if computation is ok                *
 *****************************************************************************/

int evaluate_stack(double *value,unsigned char operator, double operand)
{
    switch(operator)                                           /* decode operator */
    {
         /* identity: f(x) = x */
         case IDENT:break;                                     /* do nothing */

         /* add: f(x) = x + b */
         case ADD:  if(*value + operand > MAXDOUBLE)           /* check for positive overflow */
                       *value = MAXDOUBLE;
                    else if(*value + operand < -MAXDOUBLE)     /* check for negative overlow */
                       *value = -MAXDOUBLE;
                    else
                       *value += operand;                      /* compute result */
                    break;

         /* divide: f(x) = x / b */
         case DIV:  if(operand < 1e-300 && operand > -1e-300)  /* if divisor is nearly zero return 'invalid' */
                       return(1);
                    if(*value < 1e-300 && *value > -1e-300)    /* check for underflow */
                       break;
                    if(*value / operand > MAXDOUBLE)           /* check for positive overflow */
                       *value = MAXDOUBLE;
                    else if(*value / operand < -MAXDOUBLE)     /* check for negative overlow */
                       *value = -MAXDOUBLE;
                    else
                       *value /= operand;                      /* compute result */
                    break;

         /* subtract: f(x) = x - b */
         case SUB:  if(*value - operand > MAXDOUBLE)           /* check for positive overflow */
                       *value = MAXDOUBLE;
                    else if(*value - operand < -MAXDOUBLE)     /* check for negative overlow */
                       *value = -MAXDOUBLE;
                    else
                       *value -= operand;                      /* compute result */
                    break;

         /* multiply: f(x) = x * b */
         case MULT: if(*value < 1e-300 && *value > -1e-300)    /* check for underflow */
                    {
                        *value = 0.0;
                        break;
                    }
                    if(operand < 1e-300 && operand > -1e-300)  /* check for underflow */
                    {
                        *value = 0.0;
                        break;
                    }
                    if(*value * operand > MAXDOUBLE)           /* check for positive overflow */
                       *value = MAXDOUBLE;
                    else if(*value * operand < -MAXDOUBLE)     /* check for negative overflow */
                       *value = -MAXDOUBLE;
                    else
                       *value *= operand;
                    break;

         /* power: f(x) = x ^ b */
         case POW:  if((*value < 1e-300 && *value > -1e-300)   /* check for redundant operations (*0, *1) */
                        || *value == 1.0)
                       break;
                    if(*value < 0)                             /* internal c-function cannot use negative bases */
                       *value = -pow(-*value,(double)operand); /* compute value */
                    else
                       *value = pow(*value,(double)operand);   /* compute value */
                    break;

         /* negation: f(x) = -x */
         case NEG:  *value = -*value;                          /* compute value */
                    break;

         /* natural logarithm: f(x) = ln(x) */
         case LN:   if(*value <= 1e-300)                       /* if value almost or less than zero return 'invalid' */
                       return(1);
                    *value = log(*value);                      /* compute value */
                    break;

         /* exponential function: f(x) = e^x */
         case EXP:  if(*value > 100.0)                         /* check for overflow */
                       *value = MAXDOUBLE;
                    else
                       *value = exp(*value);                   /* compute value */
                    break;

         /* sine: f(x) = sin(x) */
         case SIN:  if(*value > 9e15 || *value < -9e15         /* if overflow would occur (cannot be handled)... */
                       || !TRIGONOMETRIC)                      /* ...or no trigonometric functions allowed... */
                       return(1);                              /* ...return 'invalid' */
                    *value = sin(*value);                      /* compute value */
                    break;

         /* cosine: f(x) = cos(x) */
         case COS:  if(*value > 9e15 || *value < -9e15         /* if overflow would occur (cannot be handled)... */
                       || !TRIGONOMETRIC)                      /* ...or no trigonometric functions allowed... */
                       return(1);                              /* ...return 'invalid' */
                    *value = cos(*value);                      /* compute value */
                    break;

         /* squareroot: f(x) = x */                           /* operation invalid if operand negative */
         case SQRT: if(*value < 0)
                       return(1);
                    *value = sqrt(*value);                     /* compute value */
                    break;

         /* reciprocal: f(x) = 1/x */
         case RECI: if(*value < 1e-300 && *value > -1e-300)    /* check for overflow and validity */
                       *value = MAXDOUBLE;
                    else
                       *value = 1.0 / *value;                  /* compute value */
                    break;

         /* operation not defined */
         default:  printf("\n\nWrong Operator Type !\n\n");
                   break;
    }
    return(0);                                                 /* formula ok */
}


/*****************************************************************************
 *   Function:    FUNCTION                                                   *
 *****************************************************************************
 *   Description: This function calculates the y-value of the function coded *
 *                by "ind" at position "x".                                  *
 *****************************************************************************
 *   Input:       CHROMOSOME *ind - the individuum to be decoded             *
 *                double x        - x-value to be inserted into function     *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      double function() - result of the computed function        *
 *****************************************************************************/

double function(CHROMOSOME *ind,double x)
{
   int i;                                /* loop index */
   double value;                         /* result of computation */

   value = x;                            /* set initial value */

   for(i=0;i<CHROMOSOME_SIZE;i+=2)       /* look at the whole chromosome */

       /* compute result of formula coded by individuum */

       if(evaluate_stack(&value,ind->coded_str[i],decode(ind->coded_str[i+1])))
       {
          /* formula is invalid */

          ind->coded_str[i] = random(IDENT+1);   /* replace invalid operation by random... */
          i-=2;                                  /* ...and try again until valid */
       }

   return(value);                                /* result of formula at ordinate x */
}


/*****************************************************************************
 *   Function:    DEST_FUNCTION                                              *
 *****************************************************************************
 *   Description: This function calculates the y-value of the destination    *
 *                function at position "x".                                  *
 *****************************************************************************
 *   Input:       double x - x-value to be inserted into function            *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      double dest_function() - result of the computed function   *
 *****************************************************************************/

double dest_function(double x)
{
    double value;

    value = 1.0 / (1.0 + exp(-x+7));    /* sigmoid function shifted by 7 to the right */
/*    value = .5 - sin(sqrt(x*x + .1))*sin(sqrt(x*x + .1)) / (1.0 + .001 * (x*x +.1) * (x*x +.1)); */
    return(value);
}



/*****************************************************************************
 *   Function:    MAIN                                                       *
 *****************************************************************************
 *   Description: This function initializes the first population and then    *
 *                calls the population module.                               *
 *****************************************************************************
 *   Input:       none                                                       *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void main()
{
  int gdriver = VGA,
      gmode = VGAHI;

  initgraph(&gdriver,&gmode,"\\BC\\BGI\\");        /* initialize graphics... */
  restorecrtmode();                                /* ...but first use textmode */

  initialize(&akt_generation);                     /* initialize first population */

  populate(&akt_generation,SELECTION_TYPE,TRIALS); /* call population module */

  closegraph();                                    /* close graphics */
}
