/*****************************************************************************
 *   Project:     GENETIC5.PRJ                                               *
 *****************************************************************************
 *   Modules:     GENPARM5.H                                                 *
 *                EXTERN5.H                                                  *
 *                INTERN5.H                                                  *
 *                GENETIC5.C                                                 *
 *                POPUL5.C                                                   *
 *                EVAL5.C                                                    *
 *                DISPLAY5.C                                                 *
 *                REPRO5.C                                                   *
 *****************************************************************************
 *   Topic:       POPUL5.C                                                   *
 *****************************************************************************
 *   Description: This module contains all functions and procedures that     *
 *                directly control the population within a GA.               *
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <mem.h>
#include "extern5.h"


/*****************************************************************************
 *   Function:    INITIALIZE                                                 *
 *****************************************************************************
 *   Description: This procedure initializes a generation on a random base.  *
 *****************************************************************************
 *   Input:       GENERATION *gen - the generation to be initialized         *
 *****************************************************************************
 *   Output:      GENERATION *gen - the initialized generation               *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void initialize(GENERATION *gen)
{
   int i,                      /* index for the population */
       j;                      /* index for the encoded alphas and betas */

   randomize();                /* set initial value for randomization */
   gen->number = 1;            /* first generation has number one */
   very_best_ind.fitness = 0;  /* set the fitness of the best individuum so far to zero */

   for(i=0;i<POP_SIZE;++i)               /* for all individuums */
       for(j=0;j<CHROMOSOME_SIZE;j+=2)    /* for all parameters */
           {
           gen->individuum[i].coded_str[j]   = random(IDENT+1);   /* encoded bytes are random */
           gen->individuum[i].coded_str[j+1] = rand();   /* encoded bytes are random */
       }

   evaluate(gen,FITNESS_TYPE);     /* compute the fitness of the first generation */
}


/*****************************************************************************
 *   Function:    ROULETTE_WHEEL                                             *
 *****************************************************************************
 *   Description: This procedure selects a parent out of one generation      *
 *                using the roulette wheel method.                           *
 *****************************************************************************
 *   Input:       GENERATION *g - the generation to be searched              *
 *                CHROMOSOME *p - the parent to be selected                  *
 *****************************************************************************
 *   Output:      CHROMOSOME *p - the parent selected                        *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void roulette_wheel(CHROMOSOME *p,GENERATION *g)
{
    double random_number;          /* random number for parent selection */
    double cumulated_fitness;      /* cumulated fitness til chromosome i */
    int i;                         /* index of selected parent           */

    /* generate a random number between zero and the total fitness of the
       generation */

    if(!str_selection[0])
           strcpy(str_selection,"roulette wheel");

        random_number = (double)rand() * g->total_fitness / (double)32767.0;

    cumulated_fitness = 0.0;       /* initialize cumulated fitness... */
    i = 0;                         /* ...and index                    */

    /* sum up the individual fitness of each chromosome until the random
       number is reached or exceeded. The chromosome that caused this is
       the selected parent */

    while((cumulated_fitness += g->individuum[i].fitness) < random_number)
         ++i;

    /* copy the information of the selected chromosome to the parent */

    memcpy(p,&g->individuum[i],sizeof(CHROMOSOME));
}



/*****************************************************************************
 *   Function:    REPLACE                                                    *
 *****************************************************************************
 *   Description: This procedure replaces one generation with another.       *
 *****************************************************************************
 *   Input:       GENERATION *g1 - the new generation (destination)          *
 *                GENERATION *g2 - the current generation (source)           *
 *                int count      - number of fittest parents to take over    *
 *                                 use 0 for generational replacement        *
 *****************************************************************************
 *   Output:      GENERATION *g1 - the new generation                        *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void replace(GENERATION *g1,GENERATION *g2,int count)
{
    int i;                                  /* loop count */
    double value;

    if(count == 0)
       memcpy(g1,g2,sizeof(GENERATION));    /* replace g1 totally with g2 */
    else
    {
        for(i=0,value = 0.0;i<POP_SIZE;++i)
        {
            /* Copy cyclically always the 'count' best individuums of the population */

            memcpy(&g1->individuum[i],&g2->individuum[sort_array[i % count].position],sizeof(CHROMOSOME));
            value += g1->individuum[i].fitness;
        }
        g1->total_fitness = value;
        g1->number = g2->number;
    }
    update_sort_array(g1);
}



/*****************************************************************************
 *   Function:    ONE_OF_THE_BEST                                            *
 *****************************************************************************
 *   Description: This procedure determines, whether an individuum of a      *
 *                generation belongs to the best for elitism.                *
 *****************************************************************************
 *   Input:       int position - index of the individuum within the          *
 *                               generation                                  *
 *                int rate     - rate, how good the individuum must be to    *
 *                               be selected for elitism                     *
 *                                  rate > 0: one of the "rate" best         *
 *                                  rate = 0: no elitism                     *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      1 - individuum belongs to the "rate" best individuums      *
 *                0 - individuum is worse than the "rate" best individuums   *
 *****************************************************************************/

int one_of_the_best(int position, int rate)
{
    int i;                                   /* loop index */

    for(i=0;i<rate;++i)                      /* look at the best individuums */
      if(sort_array[i].position == position) /* is index of individuum found in the forst positions of sort array ? */
         return(1);

    return(0);
}



/*****************************************************************************
 *   Function:    POPULATE                                                   *
 *****************************************************************************
 *   Description: This procedure is responsible for the population of the    *
 *                GA. It selects parents, recombinates and mutates them for  *
 *                a new population, evaluates the fitness and replaces the   *
 *                old generation. This is done until a given size of         *
 *                population is reached.                                     *
 *****************************************************************************
 *   Input:       GENERATION *gen              - the initial generation      *
 *                selection_type parent_select - the function that selects   *
 *                                               the parents                 *
 *                int trials                   - number of individuums to be *
 *                                               created                     *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void populate(GENERATION *gen,selection_type parent_select,long trials)
{
    long already_tried;                /* count of created individuums */
    CHROMOSOME p1,                     /* first parent to be selected  */
               p2,                     /* second parent to be selected */
                           c1,                     /* first child to be produced   */
                           c2;                     /* second child to be produced  */
        GENERATION new_generation;         /* new population               */
        int i;                             /* counts the new population    */

        already_tried = POP_SIZE;          /* one generation is already there */

        while(already_tried < trials)      /* do until desired trials are done */
        {
        for(i=0;i<POP_SIZE;i+=2)       /* you need half as many couples to be married */
        {
            parent_select(&p1,gen);    /* select the first parent */
            do
               parent_select(&p2,gen);    /* select the second parent until it is different */
            while(REPLACE_RATE != 1 && memcmp(&p1,&p2,sizeof(CHROMOSOME)) == 0);

            /* recombine the parents and get two new children */

             reproduce(&p1,&p2,&c1,&c2,CROSS_OVER_RATE,MUTATION_RATE);

            /* insert the two children into the new generation            */
            /* or copy the unmodified chromosome of the former generation */
            /* into the new generation, if it was one of the n best       */
            /* individuums (elitism)                                      */

            if(one_of_the_best(i,ELITISM_RATE))
            {
               memcpy(&new_generation.individuum[i],&gen->individuum[i],sizeof(CHROMOSOME));
               if(MODIFY_ELITIST)
                         mutate(&new_generation.individuum[i],MUTATION_RATE);
                }
                else
                   memcpy(&new_generation.individuum[i],&c1,sizeof(CHROMOSOME));

                if(one_of_the_best(i+1,ELITISM_RATE))
                {
                   memcpy(&new_generation.individuum[i+1],&gen->individuum[i+1],sizeof(CHROMOSOME));
                   if(MODIFY_ELITIST)
                         mutate(&new_generation.individuum[i+1],MUTATION_RATE);
            }
            else
               memcpy(&new_generation.individuum[i+1],&c2,sizeof(CHROMOSOME));
        }
        new_generation.number = gen->number + 1;   /* this is the next generation */

        evaluate(&new_generation,FITNESS_TYPE);     /* compute the fitness of this generation */

        replace(gen,&new_generation,REPLACE_RATE);  /* replace the current generation */

        already_tried += POP_SIZE;     /* increase the number of trials */
        }
}
