/*****************************************************************************
 *   Project:     GENETIC5.PRJ                                                *
 *****************************************************************************
 *   Modules:     GENPARM5.H                                                 *
 *                EXTERN5.H                                                  *
 *                INTERN5.H                                                  *
 *                GENETIC5.C                                                 *
 *                POPUL5.C                                                   *
 *                EVAL5.C                                                    *
 *                DISPLAY5.C                                                 *
 *                REPRO5.C                                                   *
 *****************************************************************************
 *   Topic:       REPRO5.C                                                    *
 *****************************************************************************
 *   Description: This module contains all functions and procedures that     *
 *                directly control the reproduction within a GA.             *
 *****************************************************************************/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <mem.h>
#include "extern5.h"


/*****************************************************************************
 *   Function:    CROSSOVER                                                  *
 *****************************************************************************
 *   Description: This procedure recombines two chromosomes via one point    *
 *                crossover.                                                 *
 *****************************************************************************
 *   Input:       CHROMOSOME *c1 - the first individuum to be recombined     *
 *                CHROMOSOME *c2 - the second individuum to be recombined    *
 *****************************************************************************
 *   Output:      CHROMOSOME *c1 - the first recombined individuum           *
 *                CHROMOSOME *c2 - the second recombined individuum          *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void crossover(CHROMOSOME *c1, CHROMOSOME *c2)
{
    int random_number,     /* random number for the crossover point */
        random_number2,
        i;                 /* index for the bits to be overcrossed */

    random_number = random(CHROMOSOME_SIZE); /* select random crossover point */
    random_number2 = random(CHROMOSOME_SIZE); /* select random crossover point */

    if(random_number2 > random_number)   /* random number 1 must be smaller than random number 2 */
    {
       random_number2 ^= random_number;
       random_number ^= random_number2;
       random_number2 ^= random_number;
    }

    for(i=random_number2;i<=random_number;++i)    /* All bits before the crossover point are exchanged */
    {

        c1->coded_str[i] ^= c2->coded_str[i];   /* XOR-exchange */
        c2->coded_str[i] ^= c1->coded_str[i];
        c1->coded_str[i] ^= c2->coded_str[i];
    }
}

/*****************************************************************************
 *   Function:    SET_BIT                                                    *
 *****************************************************************************
 *   Description: This procedure sets a given bit of a given value.          *
 *****************************************************************************
 *   Input:       unsigned char *value - the value to be manipulated         *
 *                int bit_no           - position of the bit to be set       *
 *****************************************************************************
 *   Output:      unsigned char *value - the manipulated value               *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void set_bit(unsigned char *value,int bit_no)
{
    unsigned char mask = 1;    /* initialize binary mask to 00000001 */

    mask <<= bit_no;           /* rotate mask "bit_no" time to the left */
    *value |= mask;            /* set bit in value via logical or */
}


/*****************************************************************************
 *   Function:    CLEAR_BIT                                                  *
 *****************************************************************************
 *   Description: This procedure clears a given bit of a given value.        *
 *****************************************************************************
 *   Input:       unsigned char *value - the value to be manipulated         *
 *                int bit_no           - position of the bit to be cleared   *
 *****************************************************************************
 *   Output:      unsigned char *value - the manipulated value               *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void clear_bit(unsigned char *value,int bit_no)
{
    unsigned char mask = 1;    /* initialize binary mask to 00000001 */

    mask <<= bit_no;           /* rotate mask "bit_no" time to the left */
    mask = ~mask;              /* negate the mask */
    *value &= mask;            /* set bit in value via logical and */
}

/*****************************************************************************
 *   Function:    MUTATE                                                     *
 *****************************************************************************
 *   Description: This procedure changes bits of a chromosome on a random    *
 *                base                                                       *
 *****************************************************************************
 *   Input:       CHROMOSOME *c - the chromosome to be mutated               *
 *                double mr     - the probability with which mutation occurs *
 *****************************************************************************
 *   Output:      CHROMOSOME *c - the mutated chromosome                     *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void mutate(CHROMOSOME *c,double mr)
{
    int i,                   /* index of the current byte of the individuum */
        j;                   /* index of the current bit of the individuum */
    double random_number;    /* random number to control the mutation */

    for(i=0;i<CHROMOSOME_SIZE;++i)    /* look at all bytes of the chromosome */
    {
        if(i & 1)
           for(j=0;j<8;++j)              /* look at every bit */
           {
              /* generate random number for mutation */
              /* note that the probability of mutation increases for less significant bits */

              random_number = (double)rand() * (4.0 / (double)(j+1)) / (double)32767.0;

              /* mutate only if random number exceeds the mutation probability */

              if(random_number < mr)
              {
                  /* generate random number for bit manipulation */

                  random_number = (double)rand() / (double)32767.0;

                  /* clear or set the bit with same probability */

                  if(random_number < 0.5)
                     set_bit(&c->coded_str[i],j);
                  else
                     clear_bit(&c->coded_str[i],j);
              }
           }
        else
        {
          random_number = (double)rand() / (double)32767.0;
          if(random_number < mr)
                c->coded_str[i] = random(IDENT+1);
        }
    }
}


/*****************************************************************************
 *   Function:    REPRODUCE                                                  *
 *****************************************************************************
 *   Description: This procedure recombines two parents to two children.     *
 *****************************************************************************
 *   Input:       CHROMOSOME *p1 - first parent                              *
 *                CHROMOSOME *p2 - second parent                             *
 *                CHROMOSOME *c1 - first child to be produced                *
 *                CHROMOSOME *c2 - second child to be produced               *
 *                double cr      - probability for crossover                 *
 *                double mr      - probability for mutation                  *
 *****************************************************************************
 *   Output:      CHROMOSOME *c1 - the first child produced                  *
 *                CHROMOSOME *c2 - the second child produced                 *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void reproduce(CHROMOSOME *p1, CHROMOSOME *p2, CHROMOSOME *c1, CHROMOSOME *c2, double cr, double mr)
{
    double random_number;             /* random number to control crossover */

    memcpy(c1,p1,sizeof(CHROMOSOME)); /* at first children and parents are identical */
    memcpy(c2,p2,sizeof(CHROMOSOME));

    /* generate random number for crossover */

    random_number = (double)rand() / (double)32767.0;

    /* do crossover, if probability is exceeded */

    if(random_number < cr)
       crossover(c1,c2);

    /* call mutation module for each child */

    mutate(c1,mr);
    mutate(c2,mr);
}
