/*  */
/* (c) 1993 by NEUROTEC GmbH 					         */
/* All rights reserved							 */
/*  */
/* Filename   : RANDOM.C                                                 */
/* Author     : Sven Feddersen						 */
/* Project    : GenLib     						 */
/*  */
/*									 */
/* Publics :								 */
/*                                                                       */
/*   double        EqualRandom(void);                                    */
/*   double        EqualDoubleRandom(double d);                          */
/*   int           EqualUnsignedRandom(unsigned u);                      */
/*   unsigned long EqualLongRandom(unsigned long l);                     */
/*   int           EqualRandomIntVec(unsigned *vec, unsigned n,          */
/*                                   unsigned mx);                       */
/*									 */
/*  */


#ifdef __TURBOC__
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#else
#include <math.h>
#include <time.h>
#include <stdlib.h>
#endif

#include <genlib.h>


extern int _GenLibError;


static boolean FirstRandom = TRUE;
static double u[98], c, cd, cm;
static int i97, j97;



/*************************************************************************/
/* This is the initialization routine for the random number generator    */
/* RANMAR(). NOTE: The seed variables can have values between:           */
/* 0 <= IJ <= 31328                                                      */
/* 0 <= KL <= 30081                                                      */
/* The random number sequences created by these two seeds are of         */
/* sufficient length to complete an entire calculation with. For         */
/* example, if several different groups are working on different parts   */
/* of the same calculation, each group could be assigned its own IJ      */
/* seed. This would leave each group with 30000 choices for the second   */
/* seed. That is to say, this random number generator can create 900     */
/* million different subsequences -- with each subsequence having a      */
/* length of approximately 10^30.                                        */
/*************************************************************************/

static void rmarin(void) {
int i, j, k, l, ii, jj, m, ij, kl;
double s, t;


  ij = rand() % 31328;
  kl = rand() % 30081;

  if (ij<0 || ij>31328 || kl<0 || kl>30081)
    exit(1);

  i = (ij / 177) % 177 + 2;
  j = ij % 177 + 2;
  k = (kl / 169) % 178 + 1;
  l = kl % 169;

  for (ii=1; ii<=97; ii++) {
    s = 0.0;
    t = 0.5;
    for (jj=1; jj<=24; jj++) {
      m = (((i * j) % 179) * k) % 179;
      i = j;
      j = k;
      k = m;
      l = (53 * l + 1) % 169;
      if ((l * m) % 64 >= 32)
	s += t;
      t *= 0.5;
    }
    u[ii] = s;
  }

  c = 362436.0 / 16777216.0;
  cd = 7654321.0 / 16777216.0;
  cm = 16777213.0 / 16777216.0;
  i97 = 97;
  j97 = 33;
}



/*************************************************************************/
/* This is the random number generator proposed by George Marsaglia in   */
/* Florida State University Report: FSU-SCRI-87-50                       */
/*************************************************************************/

static double ranmar(void) {
double uni;


  uni = u[i97] - u[j97];
  if (uni < 0.0)
    uni += 1.0;
  u[i97] = uni;
  i97--;
  if (i97==0)
    i97 = 97;
  j97--;
  if (j97==0)
    j97 = 97;
  c -= cd;
  if (c<0.0)
    c += cm;
  uni -= c;
  if (uni<0.0)
    uni += 1.0;

  return(uni);
}

/**************************************************************************/
/*  EqualRandom() liefert einen gleichverteilten Zufallswert im           */
/*  Intervall [0, 1[ mit Hilfe des C-Zufallszahlengenerators.             */
/**************************************************************************/

double EqualRandom(void) {

  if (FirstRandom) {
    srand(time(NULL));
    rmarin();
    FirstRandom = FALSE;
  }

  return(ranmar());
}



/**************************************************************************/
/*  EqualDoubleRandom() liefert einen gleichverteilten Zufallswert im     */
/*  Intervall [0, d[.                                                     */
/**************************************************************************/

double EqualDoubleRandom(double d) {

  return(d * EqualRandom());
}



/**************************************************************************/
/*  EqualUnsignedRandom() liefert einen gleichverteilten Zufallswert im   */
/*  Intervall [0, u[.                                                     */
/**************************************************************************/

int EqualUnsignedRandom(unsigned u) {

  return(u * EqualRandom());
}



/**************************************************************************/
/*  EqualLongRandom() liefert einen gleichverteilten Zufallswert im       */
/*  Intervall [0, l[.                                                     */
/**************************************************************************/

unsigned long EqualLongRandom(unsigned long l) {

  return(l * EqualRandom());
}



/**************************************************************************/
/*  EqualRandomIntVec() liefert einen Vektor von paarweise verschiedenen  */
/*  Integers.                                                             */
/*  vec : Adresse des Ergebnisvektors                                     */
/*  n   : Lnge des Ergebnisvektors                                       */
/*  mx  : Obere Grenze des Wertebereichs (Wert wird nicht angenommen)     */
/**************************************************************************/

int EqualRandomIntVec(unsigned *vec, unsigned n, unsigned mx) {
unsigned u, v;
char *set;

  set = (char *) malloc(mx);
  if (!set) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }
  memset(set, 0, mx);

  for (u=0; u<n; u++) {
    do {
      v = EqualUnsignedRandom(mx);
    } while (set[v]);
    set[v] = 1;
    vec[u] = v;
  }

  free(set);
  return(1);
}



/**************************************************************************/
/*  EqualRandom() liefert einen normalverteilten Zufallswert im           */
/*  Intervall [0, 1[.                                                     */
/**************************************************************************/

double NormalRandom(void) {

static int flag;
static double x;
double fac, r1, r2, s;

  if (FirstRandom) {
    srand(time(NULL));
    rmarin();
    FirstRandom = FALSE;
  }

  if (flag==0) {
    do {
      r1 = ranmar();
      r2 = ranmar();
      s = r1 * r1 + r2 * r2;
    } while (s >= 1.0);
    fac = sqrt(-2.0 * log(s)/s);
    x = r2 * fac;
    flag = 1;
    return(r1 * fac);
  }
  else {
    flag = 0;
    return(x);
  }
}


double NormalDoubleRandom(double d) {

  return(d * NormalRandom());
}


int NormalUnsignedRandom(unsigned u) {

  return(u * NormalRandom());
}


unsigned long NormalLongRandom(unsigned long l) {

  return(l * NormalRandom());
}
