/*  */
/* (c) 1993 by NEUROTEC GmbH 					         */
/* All rights reserved							 */
/*  */
/* Filename   : ORDMUTE.C                                                */
/* Author     : Sven Feddersen						 */
/* Project    : GenLib     						 */
/*  */
/*									 */
/* Publics :								 */
/*                                                                       */
/*   void MoveMutation(CHROM *c, unsigned p1, unsigned p2);              */
/*   void SwapMutation(CHROM *c, unsigned p1, unsigned p2);              */
/*   void InvertMutation(CHROM *c, unsigned p1, unsigned p2);            */
/*									 */
/*  */

#ifdef __TURBOC__
#include <math.h>
#include <string.h>
#include <stdlib.h>
#else
#include <math.h>
#include <string.h>
#include <stdlib.h>
#endif

#include <genlib.h>
#include <order.h>

extern int _GenLibError;

typedef void     (*MyMutationType)(CHROM*, unsigned, unsigned);

#define C_N      ((OrderRep *) c->myrep)->n
#define C_JOB    ((OrderRep *) c->myrep)->job





/*************************************************************************/
/*  MoveMutation schiebt im Individuum c den Index an der Position p2    */
/*  an die Position p1.                                                  */
/*************************************************************************/

void MoveMutation(CHROM *c, unsigned p1, unsigned p2) {
unsigned val;

  val = C_JOB[p2];
  memmove(&C_JOB[p1+1], &C_JOB[p1], (p2-p1)*sizeof(unsigned));
  C_JOB[p1] = val;
}



/*************************************************************************/
/*  SwapMutation tauscht im Individuum c die Indizes an den Positionen   */
/*  p1 und p2.                                                           */
/*************************************************************************/

void SwapMutation(CHROM *c, unsigned p1, unsigned p2) {

  SWAP(C_JOB[p1], C_JOB[p2]);
}



/*************************************************************************/
/*  InvertMutation kehrt im Individuum c die Reihenfolge der Indizes im  */
/*  Intervall [p1; p2] um.                                               */
/*************************************************************************/

void InvertMutation(CHROM *c, unsigned p1, unsigned p2) {
unsigned u, i;

  u = (p2-p1+1) / 2;
  for (i=0; i<u; i++, p1++, p2--) SWAP(C_JOB[p1], C_JOB[p2]);
}



/*************************************************************************/
/*  bin() liefert die Binomial-Wahrscheinlichkeit                        */
/*  n : Lnge der Bernoulli-Kette                                        */
/*  p : Trefferwahrscheinlichkeit                                        */
/*  k : Trefferzahl                                                      */
/*                                                                       */
/*  Hermann, D. ; "Statistik in C"; Vieweg, Braunschweig, 1991; S.34     */
/*************************************************************************/

static double bin(int n, double p, int k) {
int i, m=n+1;
double b, r, q=1.-p;

  r = log(p/q);
  b = n * log(q);
  for (i=1; i<=k; i++) b += r+log((double) m / i - 1.);
  return(exp(b));
}



/*************************************************************************/
/*  ConstantMutation fhrt eine Mutation auf einem Individuum mit einer  */
/*  vorgegebenen Mutationswahrscheinlichkeit durch.                      */
/*  mute : Mutationstyp (Funktionsadresse)                               */
/*  c    : Individuum                                                    */
/*************************************************************************/

static int ConstantMutation(MyMutationType mute, CHROM *c) {
unsigned p, n, i;
double border, rnd;
unsigned *vec;

				 /* Wie oft wird im Individuum mutiert ? */
  rnd = EqualDoubleRandom(1.0);
  border = 1.0;
  for (n=0; rnd < border && n <= C_N; n++) border -= bin(C_N, c->mutprob, n);
  n--;
  if (!n) return(1);
					     /* n paarweise verschiedene */
					      /* Zufallszahlen erzeugen  */
  vec = (unsigned *) malloc(n*sizeof(unsigned));
  if (!vec) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }
  EqualRandomIntVec(vec, n, C_N);
						 /* Mutation durchfhren */
  for (i=0; i<n; i++) {
    do p = EqualUnsignedRandom(C_N); while (p==vec[i]);
    if (vec[i] > p) SWAP(vec[i], p);
    mute(c, vec[i], p);
  }

  free(vec);
  return(1);
}






/*************************************************************************/
/*  AdaptedMutation fhrt eine Mutation auf einem Individuum mittels der */
/*  individuenspezifischen Schrittweite durch.                           */
/*  mute : Mutationstyp (Funktionsadresse)                               */
/*  c    : Individuum                                                    */
/*************************************************************************/

#define MINPROB  0.000001
#define MAXPROB  0.1
#define MINDIM  1
#define MAXDIM  (C_N / 2.0)

static int AdaptedMutation(MyMutationType mute, CHROM *c) {
double border, rnd;
unsigned m, n, p , i;
unsigned *vec;
			     /* Adaption der Mutationswahrscheinlichkeit */
  rnd = 1.0 + EqualRandom();
  if (c->mutprob == MINPROB) {
    n = EqualUnsignedRandom(2);
    if (n) c->mutprob *= rnd;
  }
  else if (c->mutprob == MAXPROB) {
    n = EqualUnsignedRandom(2);
    if (n) c->mutprob /= rnd;
  }
  else {
    n = EqualUnsignedRandom(3);
    if (!n) c->mutprob /= rnd;
    else if (n==2) c->mutprob *= rnd;
  }
  if (c->mutprob < MINPROB) c->mutprob = MINPROB;
  if (c->mutprob > MAXPROB) c->mutprob = MAXPROB;

					  /* Adaption der Mutationsweite */

  rnd = 1.0 + EqualRandom();
  if (c->mutdim == MINDIM) {
    n = EqualUnsignedRandom(2);
    if (n) c->mutdim *= rnd;
  }
  else if (c->mutdim == MAXDIM) {
    n = EqualUnsignedRandom(2);
    if (n) c->mutdim /= rnd;
  }
  else {
    n = EqualUnsignedRandom(3);
    if (!n) c->mutdim /= rnd;
    else if (n==2) c->mutdim *= rnd;
  }
  if (c->mutdim < MINDIM) c->mutdim = MINDIM;
  if (c->mutdim > MAXDIM) c->mutdim = MAXDIM;

				 /* Wie oft wird im Individuum mutiert ? */

  rnd = EqualDoubleRandom(1.0);
  border = 1.0;
  for (n=0; rnd < border && n <= C_N; n++) border -= bin(C_N, c->mutprob, n);
  n--;
  if (!n) return(1);
					     /* n paarweise verschiedene */
					     /* Zufallszahlen erzeugen   */

  vec = (unsigned *) malloc(n*sizeof(unsigned));
  if (!vec) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }
  EqualRandomIntVec(vec, n, C_N);

						 /* Mutation durchfhren */
  for (i=0; i<n; i++) {

    m = EqualUnsignedRandom((unsigned) c->mutdim);

	     /* String-Anfang und Ende drfen nicht berschritten werden */

    if (EqualUnsignedRandom(2)) p = (vec[i] + m >= C_N ? C_N - 1 : vec[i] + m);
    else p = (vec[i] < m ? 0 : vec[i]-m);

    if (vec[i] > p) SWAP(vec[i], p);
    if (vec[i] != p) mute(c, vec[i], p);
  }

  free(vec);
  return(1);
}



/*************************************************************************/
/*  ConstantMoveMutation() wendet die move mutation mit konstanter       */
/*  Mutationswahrscheinlichkeit an.                                      */
/*************************************************************************/

int ConstantMoveMutation(CHROM *c) {

  return(ConstantMutation(MoveMutation, c));
}



/*************************************************************************/
/*  ConstantSwapMutation() wendet die swap mutation mit konstanter       */
/*  Mutationswahrscheinlichkeit an.                                      */
/*************************************************************************/

int ConstantSwapMutation(CHROM *c) {

  return(ConstantMutation(SwapMutation, c));
}



/*************************************************************************/
/*  ConstantInvertMutation() wendet die invert mutation mit konstanter   */
/*  Mutationswahrscheinlichkeit an.                                      */
/*************************************************************************/

int ConstantInvertMutation(CHROM *c) {

  return(ConstantMutation(InvertMutation, c));
}



/*************************************************************************/
/*  AdaptedMoveMutation() wendet die move mutation mit adapierter        */
/*  Mutationswahrscheinlichkeit an.                                      */
/*************************************************************************/

int AdaptedMoveMutation(CHROM *c) {

  return(AdaptedMutation(MoveMutation, c));
}



/*************************************************************************/
/*  AdaptedSwapMutation() wendet die swap mutation mit adapierter        */
/*  Mutationswahrscheinlichkeit an.                                      */
/*************************************************************************/

int AdaptedSwapMutation(CHROM *c) {

  return(AdaptedMutation(SwapMutation, c));
}



/*************************************************************************/
/*  AdaptedInvertMutation() wendet die invert mutation mit adapierter    */
/*  Mutationswahrscheinlichkeit an.                                      */
/*************************************************************************/

int AdaptedInvertMutation(CHROM *c) {

  return(AdaptedMutation(InvertMutation, c));
}
