/*  */
/* (c) 1993 by NEUROTEC GmbH 					         */
/* All rights reserved							 */
/*  */
/* Filename   : ORDXOVR.C                                                */
/* Author     : Sven Feddersen						 */
/* Project    : GenLib     						 */
/*  */
/*									 */
/* Publics :								 */
/*                                                                       */
/* int EdgeRecombination(ORDER *p1, ORDER *p2, ORDER *c);                */
/* int GoldbergPMX(ORDER *p1, ORDER *p2, ORDER *c);                      */
/* int GoldbergOX(ORDER *p1, ORDER *p2, ORDER *c);                       */
/* int GoldbergCX(CHROM *p1, CHROM *p2, CHROM *c);                       */
/* int UniformOrderBased(ORDER *p1, ORDER *p2, ORDER *c);                */
/*									 */
/*  */

#include <stdlib.h>
#include <string.h>

#include <genlib.h>
#include <order.h>


#define JOBSIZE  ((OrderRep *) c->myrep)->n
#define C_N      ((OrderRep *) c->myrep)->n
#define P1_N     ((OrderRep *) p1->myrep)->n
#define P2_N     ((OrderRep *) p2->myrep)->n
#define C_JOB    ((OrderRep *) c->myrep)->job
#define P1_JOB   ((OrderRep *) p1->myrep)->job
#define P2_JOB   ((OrderRep *) p2->myrep)->job
#define HELP_JOB ((OrderRep *) help->myrep)->job



extern int _GenLibError;


typedef struct edgelist {
  unsigned jobnum;
  unsigned char numedges;
  struct edgelist *edge[4];
} edgelist;



/*************************************************************************/
/*  Edge Recombination-Operator                                          */
/*  Fr jeden Elternteil wird eine Kantenliste (Verbindungsliste)        */
/*  erzeugt. Anhand dieser Listen wird ein Kind erzeugt, das aus         */
/*  mglichst vielen Kanten der Eltenteile besteht. Sind von einem       */
/*  Knoten mehrere Wege mglich, so wird zufllig einer gewaehlt.        */
/*  Ist dagegen kein Weg ueber eine Elternkante mglich, so wird         */
/*  zufaellig einer der Knoten gewaehlt, die noch nicht durchlaufen      */
/*  wurden.                                                              */
/*************************************************************************/

int EdgeRecombination(CHROM *p1, CHROM *p2, CHROM *c) {
unsigned actjob, newjob, adr, val, i, j, k, l, num, numposs, poss[4], minlinks, startjob[2];
edgelist *list;
boolean found;
CHROM *help;

  if (P1_N != P2_N || P2_N != C_N) {
    _GenLibError = GLE_WRONG_JOBSIZE;
    return(0);
  }

  list = (edgelist *) malloc(JOBSIZE*sizeof(edgelist));
  if (!list) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }

  for (i=0; i<JOBSIZE; i++) {
    list[i].jobnum = i;
    list[i].numedges = 0;
    for (j=0; j<4; j++) list[i].edge[j] = NULL;
  }
				       /* Verzeigerung der Kantenliste */
  for (i=0; i<2; i++) {
    help = (i ? p2 : p1);
    startjob[i] = HELP_JOB[0];
    for (j=0; j<JOBSIZE; j++) {
      for (k=0; k<2; k++) {
	adr = HELP_JOB[(k ? (j==JOBSIZE-1 ? 0 : j+1) : j)];
	val = HELP_JOB[(k ? j : (j==JOBSIZE-1 ? 0 : j+1))];
	for (l=0, found=FALSE; l<list[adr].numedges && !found; l++)
	  if (list[adr].edge[l]->jobnum == val) found = TRUE;
	if (!found) list[adr].edge[list[adr].numedges++] = &list[val];
      }
    }
  }

  actjob = startjob[EqualUnsignedRandom(2)];
  for (i=0; i<JOBSIZE-1; i++) {
    C_JOB[i] = actjob;
    minlinks = 5;
    for (j=0; j<4; j++) if (list[actjob].edge[j]) {
      num = list[actjob].edge[j]->jobnum;
      if (list[num].numedges < minlinks) {
	minlinks = list[num].numedges;
	poss[0] = num;
	numposs = 1;
      }
      else if (list[num].numedges == minlinks) poss[numposs++] = num;
    }
    if (minlinks<5) newjob = poss[EqualUnsignedRandom(numposs)];
    else {
      do {
	newjob = EqualUnsignedRandom(JOBSIZE);
	for (k=0, found=FALSE; k<i+1; k++)
	  if (C_JOB[k] == newjob) found = TRUE;
      } while (found);
    }
    for (k=0; k<JOBSIZE; k++) if (k!=actjob)
      for (l=0, found=FALSE; l<4 && !found; l++)
	if (list[k].edge[l] && list[k].edge[l]->jobnum == actjob) {
	  list[k].edge[l] = NULL;
	  list[k].numedges--;
	  found = TRUE;
	}
    actjob = newjob;
  }
  C_JOB[JOBSIZE-1] = actjob;

  free(list);
  return(1);
}



/*************************************************************************/
/*  Goldberg's PMX-Operator                                              */
/*  Es wir ein zuflliges Intervall in der Reihenfolge gewhlt. In       */
/*  Intervall werden die jeweiligen Sequenzen getauscht.                 */
/*  Entsteht dabei in A' ein doppelter Wert w, so wird die Position p    */
/*  gesucht, an der w im Elternteil B steht. Danach wird wird der Wert   */
/*  w der in A' ausserhalb des Tauschintervalls liegt, durch den Wert    */
/*  ersetzt der in A an der Position p steht.                            */
/*                                                                       */
/*  PMX : A  = 8 7 3 |4 5 6 |0 2 1 9                                     */
/*        B  = 7 6 0 |1 2 9 |8 4 3 5                                     */
/*                                                                       */
/*        A' = 8 7 3 |1 2 9 |0 5 4 6                                     */
/*        B' = 7 9 0 |4 5 6 |8 1 3 2                                     */
/*************************************************************************/

int GoldbergPMX(CHROM *p1, CHROM *p2, CHROM *c) {
unsigned n1, n2, i, k;
boolean correct;

  if (P1_N != P2_N || P2_N != C_N) {
    _GenLibError = GLE_WRONG_JOBSIZE;
    return(0);
  }

				       /* Intervall whlen */
  n1 = EqualUnsignedRandom(JOBSIZE);
  do n2 = EqualUnsignedRandom(JOBSIZE); while (n2==n1);
  if (n1 > n2) SWAP(n1, n2);


				       /* Im Intervall tauschen,   */
				       /* ausserhalb kopieren bzw. */
				       /* korrigieren              */
  for (i=0; i<JOBSIZE; i++) {
    if (i<n1 || i>n2) {
      C_JOB[i] = P1_JOB[i];
      do {
	for (k=n1, correct=FALSE; k<=n2 && !correct; k++) {
	  if (C_JOB[i] == P2_JOB[k]) {
	    C_JOB[i] = P1_JOB[k];
	    correct = TRUE;
	  }
	}
      } while (correct);
    }
    else C_JOB[i] = P2_JOB[i];
  }
  return(1);
}



int GoldbergOX(CHROM *p1, CHROM *p2, CHROM *c) {
unsigned n1, n2, i, width, idx;
char *set;


  if (P1_N != P2_N || P2_N != C_N) {
    _GenLibError = GLE_WRONG_JOBSIZE;
    return(0);
  }

  set = (char *) malloc(JOBSIZE);
  if (!set) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }
  memset(set, 0, JOBSIZE);

  n1 = EqualUnsignedRandom(JOBSIZE);
  do n2 = EqualUnsignedRandom(JOBSIZE); while (n2==n1);
  if (n1 > n2) SWAP(n1, n2);

  width = n2-n1+1;

  for (idx=0; idx<width; idx++) {
    C_JOB[idx] = P2_JOB[n1+idx];
    set[C_JOB[idx]] = 1;
  }
  for (i=n1; i<JOBSIZE && idx<JOBSIZE; i++) if (!set[P1_JOB[i]]) {
    C_JOB[idx] = P1_JOB[i];
    set[C_JOB[idx]] = 1;
    idx++;
  }
  for (i=0; i<n1 && idx<JOBSIZE; i++) if (!set[P1_JOB[i]]) {
    C_JOB[idx] = P1_JOB[i];
    set[C_JOB[idx]] = 1;
    idx++;
  }

  free(set);
  return(1);
}


/*************************************************************************/
/*  Cycle crossover (CX)                                                 */
/*  Es werden hier eine Reihe von Indizes nach einem bestimmten          */
/*  Verfahren aus einem Elternteil kopiert und die verbleibenden Lcken  */
/*  mit den restlichen Indizes in der Reihenfolge gefllt, wie sie der   */
/*  zweite Elternteil vorgibt. Zur Bestimmung der zu kopierenden         */
/*  Indizes wird zunchst von einer beliebigen Position ausgegangen und  */
/*  der dort stehende Index kopiert. Der nchste zu kopierende Index     */
/*  ist dann jeweils der, der an der aktuellen Position im zweiten       */
/*  Elternteil steht. Dies wird so lange fortgefhrt, bis man wieder     */
/*  auf den ersten bearbeiteten Index trifft, sich also ein Ring         */
/*  gebildet hat.                                                        */
/*************************************************************************/

int GoldbergCX(CHROM *p1, CHROM *p2, CHROM *c) {
unsigned start, act, src, dst, i;
char *set;


  if (P1_N != P2_N || P2_N != C_N) {
    _GenLibError = GLE_WRONG_JOBSIZE;
    return(0);
  }

  set = (char *) malloc(JOBSIZE);
  if (!set) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }
  memset(set, 0, JOBSIZE);
  memset(C_JOB, 0xFF, JOBSIZE * sizeof(unsigned));

  start = EqualUnsignedRandom(JOBSIZE);
  act = start;
  do {
    C_JOB[act] = P1_JOB[act];
    set[P1_JOB[act]] = 1;
    for (i=0; i<JOBSIZE && (P1_JOB[i] != P2_JOB[act]); i++);
    act = i;
  } while (act != start);

  for (src=0, dst=0; dst<JOBSIZE;) {
    while (C_JOB[dst] != 0xFFFF && dst<JOBSIZE) dst++;
    if (dst<JOBSIZE) {
      while (set[P2_JOB[src]]) src++;
      C_JOB[dst] = P2_JOB[src];
      src++;
    }
  }

  free(set);
  return(1);
}



/*************************************************************************/
/*  Uniform order based Operator (DAVIS S.81)                            */
/*  Zu den beiden Elternteilen wird ein Bitstring gleicher Lnge         */
/*  erzeugt. Von A werden in A' alle Werte uebernommen, an deren         */
/*  Position im Bitstring eine '1' steht. Die uebrigen Werte werden in   */
/*  der Reihenfolge eingefgt, wie sie in B stehen.                      */
/*  Zur Erzeugung eines zweiten Kindes wird genau umgekehrt verfahren.   */
/*                                                                       */
/*  UOB : A  = 0 1 2 3 4 5 6 7                                           */
/*        B  = 7 5 3 1 6 4 2 0                                           */
/*             ---------------                                           */
/*             0 1 1 0 1 1 0 0                                           */
/*                                                                       */
/*        A' = 7 1 2 3 4 5 6 0                                           */
/*        B' = 7 3 4 1 5 6 2 0                                           */
/*************************************************************************/

int UniformOrderBased(CHROM *p1, CHROM *p2, CHROM *c) {
unsigned src, dst, i, j;
unsigned *p2job;


  if (P1_N != P2_N || P2_N != C_N) {
    _GenLibError = GLE_WRONG_JOBSIZE;
    return(0);
  }

  p2job = (unsigned *) malloc(JOBSIZE * sizeof(unsigned));
  if (!p2job) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }

  memcpy(p2job, P2_JOB, JOBSIZE * sizeof(unsigned));
  memset(C_JOB, 0xFF, JOBSIZE * sizeof(unsigned));

  for (i=0; i<JOBSIZE; i++) {
    if (EqualUnsignedRandom(2)) {
      C_JOB[i] = P1_JOB[i];
      for (j=0; p2job[j] != P1_JOB[i]; j++);
      p2job[j] = 0xFFFF;
    }
  }

  for (src=0, dst=0; dst<JOBSIZE;) {
    while (C_JOB[dst] != 0xFFFF && dst<JOBSIZE) dst++;
    if (dst<JOBSIZE) {
      while (p2job[src] == 0xFFFF) src++;
      C_JOB[dst] = p2job[src];
      p2job[src] = 0xFFFF;
    }
  }

  free(p2job);
  return(1);
}
